/*
 * kgraft_patch_bsc1083488
 *
 * Fix for CVE-2018-7566, bsc#1083488
 *  As a prerequisite, fix CVE-2018-1000004, bsc#1076017 as well.
 *
 *  Upstream commits:
 *  b3defb791b26 ("ALSA: seq: Make ioctls race-free")
 *  d15d662e89fc ("ALSA: seq: Fix racy pool initializations")
 *  7bd800915677 ("ALSA: seq: More protection for concurrent write and ioctl
 *                 races")
 *  d0f833065221 ("ALSA: seq: Fix possible UAF in snd_seq_check_queue()")
 *
 *  SLE12(-SP1) commits:
 *  825096fc57e7c24ada37413288883edda34187a4
 *  09a8e749f08ae3a6d33d8590925bf74bed904530
 *
 *  SLE12-SP2 commits:
 *  af154defb91ff831590e0ba20d33c08a35efdf87
 *  b00cc6647778c4639a850a9d1ad0dd2ef941ce1f
 *  be0a042cebdbfb850e20de5aa771403114e5c597
 *  a38504f5b095abf87c96d31a398f2b817fcd5086
 *  0415b4ed24a63aa1e2e4a94e95ea92b6680f69db
 *
 *  SLE12-SP3 commits:
 *  2eb8cdf7ce2b632c1cd7801b718e9611f29b28ba
 *  6aa3c5738b22d2bd414c7b610081b09bf46c86e3
 *  3086bd032f97ed10e90394edc842d3ea9212da17
 *  7c5294295914c24bd9e46d9532f98d8beb9cf2c4
 *  988351e346f9144ed3613e5743663eb35c10ce14
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/atomic.h>
#include <linux/bitmap.h>
#include <linux/list.h>
#include <linux/rwlock.h>
#include <linux/mutex.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include <asm/compat.h>
#include <sound/core.h>
#include <sound/seq_kernel.h>
#include <uapi/sound/asequencer.h>

#include "shadow.h"
#include "kgr_patch_bsc1083488.h"

#if IS_ENABLED(CONFIG_SND_SEQUENCER)

#if !IS_MODULE(CONFIG_SND_SEQUENCER)
#error "KGR patch supports only CONFIG_SND_SEQUENCER=m."
#endif

#if !IS_ENABLED(CONFIG_SMP)
#error "KGR patch supports only CONFIG_SMP=y."
#endif

#define KGR_PATCHED_MODULE "snd_seq"

static const unsigned long kgr_shadow_ioctl_mutex_id =
	KGR_SHADOW_ID(1083488, 0);


/* from sound/core/seq/seq_lock.h */
typedef atomic_t snd_use_lock_t;


struct snd_seq_event_cell;
struct snd_seq_pool;

static spinlock_t *kgr_clients_lock;
static char (*kgr_clienttablock)[SNDRV_SEQ_MAX_CLIENTS];
static struct snd_seq_client* (*kgr_clienttab)[SNDRV_SEQ_MAX_CLIENTS];

static void (*kgr_snd_use_lock_sync_helper)(snd_use_lock_t *lockp,
					    const char *file, int line);
static int (*kgr_snd_seq_delete_all_ports)(struct snd_seq_client *client);
static void (*kgr_snd_seq_queue_client_leave)(int client);
static void (*kgr_snd_seq_queue_client_termination)(int client);
static int (*kgr_snd_seq_pool_delete)(struct snd_seq_pool **ppool);
static void (*kgr_snd_seq_cell_free)(struct snd_seq_event_cell * cell);
static struct snd_seq_client_port *
(*kgr_snd_seq_port_use_ptr)(struct snd_seq_client *client,int num);
static int (*kgr_snd_seq_deliver_event)(struct snd_seq_client *client,
					struct snd_seq_event *event,
					int atomic, int hop);
static int (*kgr_snd_seq_queue_is_used)(int queueid, int client);
static int (*kgr_snd_seq_enqueue_event)(struct snd_seq_event_cell *cell,
					int atomic, int hop);
static int (*kgr_snd_seq_pool_init)(struct snd_seq_pool *pool);
static int (*kgr_check_event_type_and_length)(struct snd_seq_event *ev);
static int (*kgr_snd_seq_do_ioctl)(struct snd_seq_client *client,
				   unsigned int cmd, void __user *arg);
static int (*kgr_snd_seq_dispatch_event)(struct snd_seq_event_cell *cell,
					 int atomic, int hop);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "snd_seq:clients_lock", (void *)&kgr_clients_lock },
	{ "snd_seq:clienttablock", (void *)&kgr_clienttablock },
	{ "snd_seq:clienttab", (void *)&kgr_clienttab },
	{ "snd_seq:snd_use_lock_sync_helper",
		(void *)&kgr_snd_use_lock_sync_helper },
	{ "snd_seq:snd_seq_delete_all_ports",
		(void *)&kgr_snd_seq_delete_all_ports },
	{ "snd_seq:snd_seq_queue_client_leave",
		(void *)&kgr_snd_seq_queue_client_leave },
	{ "snd_seq:snd_seq_queue_client_termination",
		(void *)&kgr_snd_seq_queue_client_termination },
	{ "snd_seq:snd_seq_pool_delete", (void *)&kgr_snd_seq_pool_delete },
	{ "snd_seq:snd_seq_cell_free", (void *)&kgr_snd_seq_cell_free },
	{ "snd_seq:snd_seq_port_use_ptr", (void *)&kgr_snd_seq_port_use_ptr },
	{ "snd_seq:snd_seq_deliver_event", (void *)&kgr_snd_seq_deliver_event },
	{ "snd_seq:snd_seq_queue_is_used", (void *)&kgr_snd_seq_queue_is_used },
	{ "snd_seq:snd_seq_enqueue_event", (void *)&kgr_snd_seq_enqueue_event },
	{ "snd_seq:snd_seq_pool_init", (void *)&kgr_snd_seq_pool_init },
	{ "snd_seq:check_event_type_and_length",
		(void *)&kgr_check_event_type_and_length },
	{ "snd_seq:snd_seq_do_ioctl", (void *)&kgr_snd_seq_do_ioctl },
	{ "snd_seq:snd_seq_dispatch_event",
		(void *)&kgr_snd_seq_dispatch_event },
};



/* from sound/core/seq/seq_lock.h */
#define kgr_snd_use_lock_free(lockp) atomic_dec(lockp)
#define kgr_snd_use_lock_sync(lockp) kgr_snd_use_lock_sync_helper(lockp, __BASE_FILE__, __LINE__)


/* from sound/core/seq/seq_clientmgr.h */
struct snd_seq_user_client {
	struct file *file;	/* file struct of client */
	/* ... */

	/* fifo */
	struct snd_seq_fifo *fifo;	/* queue for incoming events */
	int fifo_pool_size;
};

struct snd_seq_kernel_client {
	/* ... */
};

struct snd_seq_client {
	snd_seq_client_type_t type;
	unsigned int accept_input: 1,
		accept_output: 1;
	char name[64];		/* client name */
	int number;		/* client number */
	unsigned int filter;	/* filter flags */
	DECLARE_BITMAP(event_filter, 256);
	snd_use_lock_t use_lock;
	int event_lost;
	/* ports */
	int num_ports;		/* number of ports */
	struct list_head ports_list_head;
	rwlock_t ports_lock;
	struct mutex ports_mutex;
	int convert32;		/* convert 32->64bit */

	/* output pool */
	struct snd_seq_pool *pool;		/* memory pool for this client */

	union {
		struct snd_seq_user_client user;
		struct snd_seq_kernel_client kernel;
	} data;
};


/* from sound/core/seq/seq_memory.h */
struct snd_seq_event_cell {
	struct snd_seq_event event;
	struct snd_seq_pool *pool;				/* used pool */
	struct snd_seq_event_cell *next;	/* next cell */
};

struct snd_seq_pool {
	struct snd_seq_event_cell *ptr;	/* pointer to first event chunk */
	struct snd_seq_event_cell *free;	/* pointer to the head of the free list */

	int total_elements;	/* pool size actually allocated */
	atomic_t counter;	/* cells free */

	int size;		/* pool size to be allocated */
	int room;		/* watermark for sleep/wakeup */

	int closing;

	/* statistics */
	int max_used;
	int event_alloc_nopool;
	int event_alloc_failures;
	int event_alloc_success;

	/* Write locking */
	wait_queue_head_t output_sleep;

	/* Pool lock */
	spinlock_t lock;
};

/* inlined */
static inline int kgr_snd_seq_total_cells(struct snd_seq_pool *pool)
{
	return pool ? pool->total_elements : 0;
}


/* from sound/core/seq/seq_ports.h */
struct snd_seq_port_subs_info {
	struct list_head list_head;	/* list of subscribed ports */
	unsigned int count;		/* count of subscribers */
	unsigned int exclusive: 1;	/* exclusive mode */
	struct rw_semaphore list_mutex;
	rwlock_t list_lock;
	int (*open)(void *private_data, struct snd_seq_port_subscribe *info);
	int (*close)(void *private_data, struct snd_seq_port_subscribe *info);
};

struct snd_seq_client_port {

	struct snd_seq_addr addr;	/* client/port number */
	struct module *owner;		/* owner of this port */
	char name[64];			/* port name */
	struct list_head list;		/* port list */
	snd_use_lock_t use_lock;

	/* subscribers */
	struct snd_seq_port_subs_info c_src;	/* read (sender) list */
	struct snd_seq_port_subs_info c_dest;	/* write (dest) list */

	int (*event_input)(struct snd_seq_event *ev, int direct, void *private_data,
			   int atomic, int hop);
	void (*private_free)(void *private_data);
	void *private_data;
	unsigned int callback_all : 1;
	unsigned int closing : 1;
	unsigned int timestamping: 1;
	unsigned int time_real: 1;
	int time_queue;

	/* capability, inport, output, sync */
	unsigned int capability;	/* port capability bits */
	unsigned int type;		/* port type bits */

	/* supported channels */
	int midi_channels;
	int midi_voices;
	int synth_voices;

};

#define kgr_snd_seq_port_unlock(port) kgr_snd_use_lock_free(&(port)->use_lock)


/* from sound/core/seq/seq_timer.h */
struct snd_seq_timer_tick {
	snd_seq_tick_time_t	cur_tick;	/* current tick */
	unsigned long		resolution;	/* time per tick in nsec */
	unsigned long		fraction;	/* current time per tick in nsec */
};

struct snd_seq_timer {
	/* ... tempo / offset / running state */

	unsigned int		running:1,	/* running state of queue */
				initialized:1;	/* timer is initialized */

	unsigned int		tempo;		/* current tempo, us/tick */
	int			ppq;		/* time resolution, ticks/quarter */

	snd_seq_real_time_t	cur_time;	/* current time */
	struct snd_seq_timer_tick	tick;	/* current tick */
	int tick_updated;

	int			type;		/* timer type */
	struct snd_timer_id	alsa_id;	/* ALSA's timer ID */
	struct snd_timer_instance	*timeri;	/* timer instance */
	unsigned int		ticks;
	unsigned long		preferred_resolution; /* timer resolution, ticks/sec */

	unsigned int skew;
	unsigned int skew_base;

	struct timeval		last_update;	 /* time of last clock update, used for interpolation */

	spinlock_t lock;
};

/* inlined */
static inline int kgr_snd_seq_compare_tick_time(snd_seq_tick_time_t *a, snd_seq_tick_time_t *b)
{
	/* compare ticks */
	return (*a >= *b);
}

/* inlined */
static inline int kgr_snd_seq_compare_real_time(snd_seq_real_time_t *a, snd_seq_real_time_t *b)
{
	/* compare real time */
	if (a->tv_sec > b->tv_sec)
		return 1;
	if ((a->tv_sec == b->tv_sec) && (a->tv_nsec >= b->tv_nsec))
		return 1;
	return 0;
}


/* from sound/core/seq/seq_prioq.h */
struct snd_seq_prioq {
	struct snd_seq_event_cell *head;      /* pointer to head of prioq */
	struct snd_seq_event_cell *tail;      /* pointer to tail of prioq */
	int cells;
	spinlock_t lock;
};


/* from sound/core/seq/seq_queue.h */
struct snd_seq_queue {
	int queue;		/* queue number */

	char name[64];		/* name of this queue */

	struct snd_seq_prioq	*tickq;		/* midi tick event queue */
	struct snd_seq_prioq	*timeq;		/* real-time event queue */

	struct snd_seq_timer *timer;	/* time keeper for this queue */
	int	owner;		/* client that 'owns' the timer */
	unsigned int	locked:1,	/* timer is only accesibble by owner if set */
		klocked:1,	/* kernel lock (after START) */
		check_again:1,
		check_blocked:1;

	unsigned int flags;		/* status flags */
	unsigned int info_flags;	/* info for sync */

	spinlock_t owner_lock;
	spinlock_t check_lock;

	/* clients which uses this queue (bitmap) */
	DECLARE_BITMAP(clients_bitmap, SNDRV_SEQ_MAX_CLIENTS);
	unsigned int clients;	/* users of this queue */
	struct mutex timer_mutex;

	snd_use_lock_t use_lock;
};



/* from sound/core/seq/seq_clientmgr.c */
#define SNDRV_SEQ_LFLG_INPUT	0x0001
#define SNDRV_SEQ_LFLG_OUTPUT	0x0002
#define SNDRV_SEQ_LFLG_OPEN	(SNDRV_SEQ_LFLG_INPUT|SNDRV_SEQ_LFLG_OUTPUT)

/* inlined */
static inline unsigned short kgr_snd_seq_file_flags(struct file *file)
{
	switch (file->f_mode & (FMODE_READ | FMODE_WRITE)) {
	case FMODE_WRITE:
		return SNDRV_SEQ_LFLG_OUTPUT;
	case FMODE_READ:
		return SNDRV_SEQ_LFLG_INPUT;
	default:
		return SNDRV_SEQ_LFLG_OPEN;
	}
}

/* inlined */
static inline int
kgr_snd_seq_write_pool_allocated(struct snd_seq_client *client)
{
	return kgr_snd_seq_total_cells(client->pool) > 0;
}



/* New */
static int __kgr_shadow_ioctl_mutex_init(void *new_shadow_data,
					 void *init_data)
{
	struct mutex *m = new_shadow_data;

	mutex_init(m);
	return 0;
}

/* New */
static struct mutex* kgr_get_shadow_ioctl_mutex(struct snd_seq_client *client)
{
	return klp_shadow_get_or_alloc_with_init(client,
					kgr_shadow_ioctl_mutex_id,
					__kgr_shadow_ioctl_mutex_init, NULL,
					sizeof(struct mutex), GFP_KERNEL);
}

/* Patched */
int kgr_seq_free_client1(struct snd_seq_client *client)
{
	unsigned long flags;

	if (!client)
		return 0;
	kgr_snd_seq_delete_all_ports(client);
	kgr_snd_seq_queue_client_leave(client->number);
	spin_lock_irqsave(kgr_clients_lock, flags);
	(*kgr_clienttablock)[client->number] = 1;
	(*kgr_clienttab)[client->number] = NULL;
	spin_unlock_irqrestore(kgr_clients_lock, flags);
	kgr_snd_use_lock_sync(&client->use_lock);
	kgr_snd_seq_queue_client_termination(client->number);
	if (client->pool)
		kgr_snd_seq_pool_delete(&client->pool);
	spin_lock_irqsave(kgr_clients_lock, flags);
	(*kgr_clienttablock)[client->number] = 0;
	spin_unlock_irqrestore(kgr_clients_lock, flags);
	/*
	 * Fix CVE-2018-1000004
	 *  +1 line
	 */
	klp_shadow_free(client, kgr_shadow_ioctl_mutex_id);
	return 0;
}

/* Patched, callers also patched */
static int kgr_snd_seq_cell_alloc(struct snd_seq_pool *pool,
				  struct snd_seq_event_cell **cellp,
				  /*
				   * Fix CVE-2018-7566
				   *  -1 line, +2 lines
				   */
				  int nonblock, struct file *file,
				  struct mutex *mutexp)
{
	struct snd_seq_event_cell *cell;
	unsigned long flags;
	int err = -EAGAIN;
	wait_queue_t wait;

	if (pool == NULL)
		return -EINVAL;

	*cellp = NULL;

	init_waitqueue_entry(&wait, current);
	spin_lock_irqsave(&pool->lock, flags);
	if (pool->ptr == NULL) {	/* not initialized */
		pr_debug("ALSA: seq: pool is not initialized\n");
		err = -EINVAL;
		goto __error;
	}
	while (pool->free == NULL && ! nonblock && ! pool->closing) {

		set_current_state(TASK_INTERRUPTIBLE);
		add_wait_queue(&pool->output_sleep, &wait);
		spin_unlock_irq(&pool->lock);
		/*
		 * Fix CVE-2018-7566
		 *  +2 lines
		 */
		if (mutexp)
			mutex_unlock(mutexp);
		schedule();
		/*
		 * Fix CVE-2018-7566
		 *  +2 lines
		 */
		if (mutexp)
			mutex_lock(mutexp);
		spin_lock_irq(&pool->lock);
		remove_wait_queue(&pool->output_sleep, &wait);
		/* interrupted? */
		if (signal_pending(current)) {
			err = -ERESTARTSYS;
			goto __error;
		}
	}
	if (pool->closing) { /* closing.. */
		err = -ENOMEM;
		goto __error;
	}

	cell = pool->free;
	if (cell) {
		int used;
		pool->free = cell->next;
		atomic_inc(&pool->counter);
		used = atomic_read(&pool->counter);
		if (pool->max_used < used)
			pool->max_used = used;
		pool->event_alloc_success++;
		/* clear cell pointers */
		cell->next = NULL;
		err = 0;
	} else
		pool->event_alloc_failures++;
	*cellp = cell;

__error:
	spin_unlock_irqrestore(&pool->lock, flags);
	return err;
}

/* Patched, callers passing a non-NULL mutexp also patched */
static int
kgr_snd_seq_event_dup(struct snd_seq_pool *pool, struct snd_seq_event *event,
		      struct snd_seq_event_cell **cellp, int nonblock,
		      /*
		       * Fix CVE-2018-7566
		       *  -1 line, +1 line
		       */
		      struct file *file, struct mutex *mutexp)
{
	int ncells, err;
	unsigned int extlen;
	struct snd_seq_event_cell *cell;

	*cellp = NULL;

	ncells = 0;
	extlen = 0;
	if (snd_seq_ev_is_variable(event)) {
		extlen = event->data.ext.len & ~SNDRV_SEQ_EXT_MASK;
		ncells = (extlen + sizeof(struct snd_seq_event) - 1) / sizeof(struct snd_seq_event);
	}
	if (ncells >= pool->total_elements)
		return -ENOMEM;

	/*
	 * Fix CVE-2018-7566
	 *  -1 line, +1 line
	 */
	err = kgr_snd_seq_cell_alloc(pool, &cell, nonblock, file, mutexp);
	if (err < 0)
		return err;

	/* copy the event */
	cell->event = *event;

	/* decompose */
	if (snd_seq_ev_is_variable(event)) {
		int len = extlen;
		int is_chained = event->data.ext.len & SNDRV_SEQ_EXT_CHAINED;
		int is_usrptr = event->data.ext.len & SNDRV_SEQ_EXT_USRPTR;
		struct snd_seq_event_cell *src, *tmp, *tail;
		char *buf;

		cell->event.data.ext.len = extlen | SNDRV_SEQ_EXT_CHAINED;
		cell->event.data.ext.ptr = NULL;

		src = (struct snd_seq_event_cell *)event->data.ext.ptr;
		buf = (char *)event->data.ext.ptr;
		tail = NULL;

		while (ncells-- > 0) {
			int size = sizeof(struct snd_seq_event);
			if (len < size)
				size = len;
			/*
			 * Fix CVE-2018-7566
			 *  -1 line, +2 lines
			 */
			err = kgr_snd_seq_cell_alloc(pool, &tmp, nonblock, file,
						     mutexp);
			if (err < 0)
				goto __error;
			if (cell->event.data.ext.ptr == NULL)
				cell->event.data.ext.ptr = tmp;
			if (tail)
				tail->next = tmp;
			tail = tmp;
			/* copy chunk */
			if (is_chained && src) {
				tmp->event = src->event;
				src = src->next;
			} else if (is_usrptr) {
				if (copy_from_user(&tmp->event, (char __force __user *)buf, size)) {
					err = -EFAULT;
					goto __error;
				}
			} else {
				memcpy(&tmp->event, buf, size);
			}
			buf += size;
			len -= size;
		}
	}

	*cellp = cell;
	return 0;

__error:
	kgr_snd_seq_cell_free(cell);
	return err;
}

/* Patched, callers passing a non-NULL mutexp also patched */
static int kgr_snd_seq_client_enqueue_event(struct snd_seq_client *client,
					    struct snd_seq_event *event,
					    struct file *file, int blocking,
					    /*
					     * Fix CVE-2018-7566
					     *  -1 line, +2 lines
					     */
					    int atomic, int hop,
					    struct mutex *mutexp)
{
	struct snd_seq_event_cell *cell;
	int err;

	/* special queue values - force direct passing */
	if (event->queue == SNDRV_SEQ_ADDRESS_SUBSCRIBERS) {
		event->dest.client = SNDRV_SEQ_ADDRESS_SUBSCRIBERS;
		event->queue = SNDRV_SEQ_QUEUE_DIRECT;
	} else
#ifdef SUPPORT_BROADCAST
		if (event->queue == SNDRV_SEQ_ADDRESS_BROADCAST) {
			event->dest.client = SNDRV_SEQ_ADDRESS_BROADCAST;
			event->queue = SNDRV_SEQ_QUEUE_DIRECT;
		}
#endif
	if (event->dest.client == SNDRV_SEQ_ADDRESS_SUBSCRIBERS) {
		/* check presence of source port */
		struct snd_seq_client_port *src_port = kgr_snd_seq_port_use_ptr(client, event->source.port);
		if (src_port == NULL)
			return -EINVAL;
		kgr_snd_seq_port_unlock(src_port);
	}

	/* direct event processing without enqueued */
	if (snd_seq_ev_is_direct(event)) {
		if (event->type == SNDRV_SEQ_EVENT_NOTE)
			return -EINVAL; /* this event must be enqueued! */
		return kgr_snd_seq_deliver_event(client, event, atomic, hop);
	}

	/* Not direct, normal queuing */
	if (kgr_snd_seq_queue_is_used(event->queue, client->number) <= 0)
		return -EINVAL;  /* invalid queue */
	if (! kgr_snd_seq_write_pool_allocated(client))
		return -ENXIO; /* queue is not allocated */

	/* allocate an event cell */
	/*
	 * Fix CVE-2018-7566
	 *  -1 line, +2 lines
	 */
	err = kgr_snd_seq_event_dup(client->pool, event, &cell, !blocking || atomic,
				    file, mutexp);
	if (err < 0)
		return err;

	/* we got a cell. enqueue it. */
	if ((err = kgr_snd_seq_enqueue_event(cell, atomic, hop)) < 0) {
		kgr_snd_seq_cell_free(cell);
		return err;
	}

	return 0;
}

/* Patched */
ssize_t kgr_snd_seq_write(struct file *file, const char __user *buf,
			  size_t count, loff_t *offset)
{
	struct snd_seq_client *client = file->private_data;
	int written = 0, len;
	/*
	 * Fix CVE-2018-7566
	 *  -1 line, +1 line
	 */
	int err;
	struct snd_seq_event event;
	/*
	 * Fix CVE-2018-7566
	 *  +1 line
	 */
	struct mutex *shadow_ioctl_mutex;

	if (!(kgr_snd_seq_file_flags(file) & SNDRV_SEQ_LFLG_OUTPUT))
		return -ENXIO;

	/* check client structures are in place */
	if (snd_BUG_ON(!client))
		return -ENXIO;

	if (!client->accept_output || client->pool == NULL)
		return -ENXIO;

	/* allocate the pool now if the pool is not allocated yet */
	/*
	 * Fix CVE-2018-7566
	 *  +4 lines
	 */
	shadow_ioctl_mutex = kgr_get_shadow_ioctl_mutex(client);
	if (!shadow_ioctl_mutex)
		return -ENOMEM;
	mutex_lock(shadow_ioctl_mutex);
	if (client->pool->size > 0 && !kgr_snd_seq_write_pool_allocated(client)) {
		/*
		 * Fix CVE-2018-7566
		 *  -2 lines, +3 lines
		 */
		err = kgr_snd_seq_pool_init(client->pool);
		if (err < 0)
			goto out;
	}

	/* only process whole events */
	/*
	 * Fix CVE-2018-7566
	 *  +1 line
	 */
	err = -EINVAL;
	while (count >= sizeof(struct snd_seq_event)) {
		/* Read in the event header from the user */
		len = sizeof(event);
		if (copy_from_user(&event, buf, len)) {
			err = -EFAULT;
			break;
		}
		event.source.client = client->number;	/* fill in client number */
		/* Check for extension data length */
		if (kgr_check_event_type_and_length(&event)) {
			err = -EINVAL;
			break;
		}

		/* check for special events */
		if (event.type == SNDRV_SEQ_EVENT_NONE)
			goto __skip_event;
		else if (snd_seq_ev_is_reserved(&event)) {
			err = -EINVAL;
			break;
		}

		if (snd_seq_ev_is_variable(&event)) {
			int extlen = event.data.ext.len & ~SNDRV_SEQ_EXT_MASK;
			if ((size_t)(extlen + len) > count) {
				/* back out, will get an error this time or next */
				err = -EINVAL;
				break;
			}
			/* set user space pointer */
			event.data.ext.len = extlen | SNDRV_SEQ_EXT_USRPTR;
			event.data.ext.ptr = (char __force *)buf
						+ sizeof(struct snd_seq_event);
			len += extlen; /* increment data length */
		} else {
#ifdef CONFIG_COMPAT
			if (client->convert32 && snd_seq_ev_is_varusr(&event)) {
				void *ptr = (void __force *)compat_ptr(event.data.raw32.d[1]);
				event.data.ext.ptr = ptr;
			}
#endif
		}

		/* ok, enqueue it */
		err = kgr_snd_seq_client_enqueue_event(client, &event, file,
						       !(file->f_flags & O_NONBLOCK),
							/*
							 * Fix CVE-2018-7566
							 *  -1 line, +1 line
							 */
						       0, 0, shadow_ioctl_mutex);
		if (err < 0)
			break;

	__skip_event:
		/* Update pointers and counts */
		count -= len;
		buf += len;
		written += len;
	}

	/*
	 * Fix CVE-2018-7566
	 *  +2 lines
	 */
out:
	mutex_unlock(shadow_ioctl_mutex);
	return written ? written : err;
}

/* Patched */
long kgr_snd_seq_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	struct snd_seq_client *client = file->private_data;
	/*
	 * Fix CVE-2018-1000004
	 *  +2 lines
	 */
	long ret;
	struct mutex *shadow_ioctl_mutex;

	if (snd_BUG_ON(!client))
		return -ENXIO;

	/*
	 * Fix CVE-2018-1000004
	 *  -1 line, +7 lines
	 */
	shadow_ioctl_mutex = kgr_get_shadow_ioctl_mutex(client);
	if (!shadow_ioctl_mutex)
		return -ENOMEM;
	mutex_lock(shadow_ioctl_mutex);
	ret = kgr_snd_seq_do_ioctl(client, cmd, (void __user *) arg);
	mutex_unlock(shadow_ioctl_mutex);
	return ret;
}


/* New */
static int kgr_event_is_ready(struct snd_seq_event *ev, void *current_time)
{
	if ((ev->flags & SNDRV_SEQ_TIME_STAMP_MASK) == SNDRV_SEQ_TIME_STAMP_TICK)
		return kgr_snd_seq_compare_tick_time(current_time, &ev->time.tick);
	else
		return kgr_snd_seq_compare_real_time(current_time, &ev->time.time);
}

/* Patched, all callers passing in a non-NULL current_time also patched */
static
/*
 * Fix CVE-2018-7566
 *  -1 line, +2 lines
 */
struct snd_seq_event_cell *kgr_snd_seq_prioq_cell_out(struct snd_seq_prioq *f,
						      void *current_time)
{
	struct snd_seq_event_cell *cell;
	unsigned long flags;

	if (f == NULL) {
		pr_debug("ALSA: seq: snd_seq_prioq_cell_in() called with NULL prioq\n");
		return NULL;
	}
	spin_lock_irqsave(&f->lock, flags);

	cell = f->head;
	/*
	 * Fix CVE-2018-7566
	 *  +2 lines
	 */
	if (cell && current_time && !kgr_event_is_ready(&cell->event, current_time))
		cell = NULL;
	if (cell) {
		f->head = cell->next;

		/* reset tail if this was the last element */
		if (f->tail == cell)
			f->tail = NULL;

		cell->next = NULL;
		f->cells--;
	}

	spin_unlock_irqrestore(&f->lock, flags);
	return cell;
}

/* Patched */
void kgr_snd_seq_check_queue(struct snd_seq_queue *q, int atomic, int hop)
{
	unsigned long flags;
	struct snd_seq_event_cell *cell;

	if (q == NULL)
		return;

	/* make this function non-reentrant */
	spin_lock_irqsave(&q->check_lock, flags);
	if (q->check_blocked) {
		q->check_again = 1;
		spin_unlock_irqrestore(&q->check_lock, flags);
		return;		/* other thread is already checking queues */
	}
	q->check_blocked = 1;
	spin_unlock_irqrestore(&q->check_lock, flags);

      __again:
	/* Process tick queue... */
	/*
	 * Fix CVE-2018-7566
	 *  -11 line, +7 lines
	 */
	for (;;) {
		cell = kgr_snd_seq_prioq_cell_out(q->tickq,
						  &q->timer->tick.cur_tick);
		if (!cell)
			break;
		kgr_snd_seq_dispatch_event(cell, atomic, hop);
	}

	/* Process time queue... */
	/*
	 * Fix CVE-2018-7566
	 *  -11 line, +7 lines
	 */
	for (;;) {
		cell = kgr_snd_seq_prioq_cell_out(q->timeq,
						  &q->timer->cur_time);
		if (!cell)
			break;
		kgr_snd_seq_dispatch_event(cell, atomic, hop);
	}

	/* free lock */
	spin_lock_irqsave(&q->check_lock, flags);
	if (q->check_again) {
		q->check_again = 0;
		spin_unlock_irqrestore(&q->check_lock, flags);
		goto __again;
	}
	q->check_blocked = 0;
	spin_unlock_irqrestore(&q->check_lock, flags);
}



static int kgr_patch_bsc1083488_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1083488_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1083488_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1083488_module_nb = {
	.notifier_call = kgr_patch_bsc1083488_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1083488_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1083488_kallsyms();
		if (ret) {
			mutex_unlock(&module_mutex);
			return ret;
		}
	}

	ret = register_module_notifier(&kgr_patch_bsc1083488_module_nb);
	mutex_unlock(&module_mutex);
	if (ret)
		return ret;

	ret = kgr_shadow_init();
	if (ret)
		unregister_module_notifier(&kgr_patch_bsc1083488_module_nb);

	return ret;
}

void kgr_patch_bsc1083488_cleanup(void)
{
	kgr_shadow_cleanup();
	unregister_module_notifier(&kgr_patch_bsc1083488_module_nb);
}

#endif  /* IS_ENABLED(CONFIG_SND_SEQUENCER) */
